/*
 Xye License (it is a PNG/ZLIB license)

Copyright (c) 2006 Victor Hugo Soliz Kuncar

This software is provided 'as-is', without any express or implied warranty. In no event will the authors be held liable for any damages arising from the use of this software.

Permission is granted to anyone to use this software for any purpose, including commercial applications, and to alter it and redistribute it freely, subject to the following restrictions:

    1. The origin of this software must not be misrepresented; you must not claim that you wrote the original software. If you use this software in a product, an acknowledgment in the product documentation would be appreciated but is not required.

    2. Altered source versions must be plainly marked as such, and must not be misrepresented as being the original software.

    3. This notice may not be removed or altered from any source distribution.

*/

#include "font.h"
#include <stdexcept>
#include "gen.h"

static Uint32 GetPixel(SDL_Surface *Surface, Sint32 X, Sint32 Y)
{
   Uint8  *bits;
   Uint32 Bpp;

   X=(X>=0)?X:-X;
   X=(X<Surface->w)?X:Surface->w;

   Bpp = Surface->format->BytesPerPixel;
   bits = ((Uint8 *)Surface->pixels)+Y*Surface->pitch+X*Bpp;

   // Get the pixel
   switch(Bpp) {
      case 1:
         return *((Uint8 *)Surface->pixels + Y * Surface->pitch + X);
         break;
      case 2:
         return *((Uint16 *)Surface->pixels + Y * Surface->pitch/2 + X);
         break;
      case 3: { // Format/endian independent
         Uint8 r, g, b;
         r = *((bits)+Surface->format->Rshift/8);
         g = *((bits)+Surface->format->Gshift/8);
         b = *((bits)+Surface->format->Bshift/8);
         return SDL_MapRGB(Surface->format, r, g, b);
         }
         break;
      case 4:
         return *((Uint32 *)Surface->pixels + Y * Surface->pitch/4 + X);
         break;
   }

   return 0xF0000000;
}

Font::Font(SDL_Surface* surface, Uint8 R,  Uint8 G, Uint8 B)
{

    int x = 0, i = 0, j;
    Uint32 pixel;
    Uint32 pink;
    if (surface==NULL)
    {
        fprintf(stderr,"Couldn't initialize font.");
        throw ("Couldn't initialize font.");
    }

    int w=surface->w,h=surface->h;


    tex = CreateFixedRGBASurface(0, w,h);
    SDL_LockSurface(surface);
    SDL_LockSurface(tex);


    SDL_PixelFormat* surform=surface->format;

    pink = SDL_MapRGB(surform, 255, 0, 255);
    while (x < w)
    {
        if (GetPixel(surface, x, 0) == pink)
        {
    	    CharPos[i++]=x;
    	    while((x < w) && (GetPixel(surface, x, 0)== pink)) x++;
            CharPos[i++]=x;
        }
        x++;
    }
    MaxPos = x-1;
    Uint8 rd=0,g=0,b=0,a=0;
    FixedRGBA c;
    c.r=R;
    c.g=G;
    c.b=B;
    c.a=175;
    Uint32* c_Uint32 = (Uint32*)(&c);

    FixedRGBA* X;


    Uint8* Calpha= &(c.a);

    Uint32* pixels = (Uint32*)(tex->pixels);

    for (i=0;i<w;i++)
    for (j=0;j<h;j++)
    {
    	SDL_GetRGB(GetPixel(surface,i,j), surform, &rd, Calpha, &b);
        *(pixels + (j * w) +  (i)) = *c_Uint32;
    }

    SDL_UnlockSurface(surface);
    SDL_UnlockSurface(tex);
}


Font::Font(SDL_Surface* surface, SDL_Color& color)
{
	Font(surface, color.r, color.g, color.b);
}

Font::~Font()
{
    SDL_FreeSurface(tex);
}


void Font::WriteWrap(SDL_Surface* surface, int x, int y, int w, int h,const std::string& text) const
{
	char* c=string2charp(  (string*)(&text) );
    WriteWrap(surface,x,y,w,h,c);
    delete[] c;
}


void Font::Write(SDL_Surface *Surface, int x, int y, const std::string& text) const
{
	char* c=string2charp(  (string*)(&text) );
    Write(Surface,x,y,c);
    delete[] c;
}


void Font::Write(SDL_Surface *Surface, int x, int y, const char *text) const
{
    const char* c;
    int charoffset;
    unsigned int fh=Height();
    SDL_Rect srcrect, dstrect;

    if(text == NULL)
	return;

    // these values won't change in the loop
    srcrect.y = 1;
    dstrect.y = y;
    srcrect.h = dstrect.h = tex->h - 1;

    Sint16 sw=Surface->w;
    Sint16 sh=Surface->h;
    int ox=x;
    int tw;


    for(c = text; (*c != '\0') && (x <= sw) && (y<=sh) ; c++)
    {
        charoffset = ((int) (*c - 33)) * 2 + 1;
        // skip spaces and nonprintable characters

        if (*c == '\n')
        {
            x=ox;
            y+=fh;
            dstrect.y=y;
            continue;
        }

        if (*c == ' ' || charoffset < 0 || charoffset > MaxPos) {
            x += CharPos[2]-CharPos[1];
            continue;
        }

         tw=srcrect.w = dstrect.w =
            (CharPos[charoffset+2] + CharPos[charoffset+1])/2 -
            (CharPos[charoffset] + CharPos[charoffset-1])/2;


        srcrect.x = (CharPos[charoffset]+CharPos[charoffset-1])/2;
        dstrect.x = int(x - (CharPos[charoffset] - CharPos[charoffset-1]) /2.0);

        if (x+tw>=0) SDL_BlitSurface(tex, &srcrect, Surface, &dstrect);

        x += CharPos[charoffset+1] - CharPos[charoffset];
    }
}

bool IsNormalChar(const char &a)
{
    return (((a>='0') && (a<='9')) || ((a>='a') && (a<='z')) || ((a>='A') && (a<='Z')));
}


void Font::WriteWrap(SDL_Surface* Surface, int x, int y, int w, int h,const char* text) const
{
	char word[21];
	int wi,autoprint;
    const char* c;
    int charoffset;
    unsigned int fh=Height();
    SDL_Rect srcrect, dstrect;

    if(text == NULL)
	return;

    // these values won't change in the loop
    srcrect.y = 1;
    dstrect.y = y;
    dstrect.h = fh - 1;

    Sint16 sw=Surface->w;
    Sint16 sh=Surface->h;
    sw= sw<x+w?sw:(x+w);
    sh= sh<y+h?sh:(y+h);
    int ox=x;
    int tw;
    srcrect.h = (sh-y<fh)?sh-y:fh;


    word[0]='\0';
    wi=autoprint=0;
    for(c = text; (y<=sh) ; c++)
    {
        //fprintf(stderr,"hoo %c %d\n",*c,autoprint);
        if (autoprint>0)
        {
            if (*c=='\0') break;
            autoprint--;

            charoffset = ((int) (*c - 33)) * 2 + 1;

            if (*c == '\n')
            {
                x=ox;
                y+=fh;
                dstrect.y=y;
                srcrect.h = (sh-y<fh)?sh-y:fh;
                continue;
            }
            // skip spaces and nonprintable characters
            if (*c == ' ' || charoffset < 0 || charoffset > MaxPos) {
                x += CharPos[2]-CharPos[1];
                continue;
            }
            //if (x <= sw) break; //too big
             tw=srcrect.w = dstrect.w =
                (CharPos[charoffset+2] + CharPos[charoffset+1])/2 -
                (CharPos[charoffset] + CharPos[charoffset-1])/2;


            srcrect.x = (CharPos[charoffset]+CharPos[charoffset-1])/2;
            dstrect.x = int(x - (CharPos[charoffset] - CharPos[charoffset-1]) /2.0);

            if (x+tw>=sw)
            {
                dstrect.x=x=ox;
                y+=fh;
                dstrect.y=y;
                srcrect.h = (sh-y<fh)?sh-y:fh;
                if (y>sh) continue;
            }
            x += CharPos[charoffset+1] - CharPos[charoffset];
            if (x+tw>=0) SDL_BlitSurface(tex, &srcrect, Surface, &dstrect);
        }
        else if ( (IsNormalChar(*c)) && (wi<20))
        {
            word[wi]=*c;
            wi++;
            word[wi]='\0';
        }
        else
        {
            if (x+TextWidth(word)+1>=sw)
            {
                dstrect.x=x=ox;
                y+=fh;
                dstrect.y=y;
                srcrect.h = (sh-y<fh)?sh-y:fh;

                //if (y>sh)
                  //  continue;
            }
            autoprint=wi+1;
            c=c-wi-1;
            wi=0;
            word[0]='\0';

        }



    }


}









unsigned int Font::TextWidth(const char *text) const
{
    const char* c;
    int charoffset=0;
    unsigned int width = 0;

    if(text == NULL)
	return 0;

    for(c = text; *c != '\0'; c++) {
        charoffset = ((int) *c - 33) * 2 + 1;
        // skip spaces and nonprintable characters
        if (*c == ' ' || charoffset < 0 || charoffset > MaxPos) {
            width += CharPos[2]-CharPos[1];
            continue;
        }

        width += CharPos[charoffset+1] - CharPos[charoffset];
    }

    return width;

}

unsigned int Font::Height() const
{
    return tex->h - 1;
}


void Font::WriteCenter(SDL_Surface *Surface, int y, const char *text)
{
	Write(Surface, Surface->w/2 - TextWidth(text)/2, y, text);
}
